(in-package "ACL2")

#|

Briefly, negative-syntaxp recognizes terms which are "syntactically negative."  These are terms which could
be simplified by multiplying them by -1.  For example, the term
  (integerp (+ -1/2 (* -1 x)))
could be simplified by multiplying it by -1, yielding (with distributivity enabled) the term
  (integerp (+ 1/2 x))
. Often, a statement about a negative term can be simplified by rewriting it into a statement about the term's
positive analogue.  For example, (integerp (* -1 x)) can be simplified to (integerp x), if x is an
acl2-numberp.  However, a rule rewriting (integerp (* -1 x)) to (integerp x) is to general enough to catch all
"negative" terms.  For example, the rule would not fire on (integerp (- x)).  negative-syntaxp provides a
general and extensible facility for making such rules.

Note that the notion of "syntactically negative" pays no attention to *values*. E.g., (- x) is "syntactically negative" regardless of the sign of the value of x.

See also factor-2.lisp for similar functions which tell whether a term could be simplified by multiplying (or
dividing) it by 2.

|#



;this function operates on translated terms!
;this function should be extended to recognize more terms.
;do I need to test that binary-+ and binary-* have the right number of arguments?
(defun negative-syntaxp (term)
  (if (not (consp term))
      nil
    (case (car term)
      (quote (and (rationalp (cadr term)) 
                  (< (cadr term) 0)))
      (unary-- (not (negative-syntaxp (cadr term))))
      (binary-+ (and ;(equal (length term) 3) ;ok since binary-+ should always have 2 args
                     (negative-syntaxp (cadr term))
                     (negative-syntaxp (caddr term))))
      (binary-* (and ;(equal (length term) 3)
                     (or (and (negative-syntaxp (cadr term))
                              (not (negative-syntaxp (caddr term))))
                         (and (not (negative-syntaxp (cadr term)))
                              (negative-syntaxp (caddr term)))))))))


#|

The following terms are "negative":
 negative constants
 (- x)
 (* -1 x)
 (* -5/7 x)
 (/ (- a)) - how do we simp this?
 (+ -1/2 (* -1/2 x))
 the sum of two (or more) negative terms


;list is a true list of terms



neg
 negative numeric constants
 (- <non-neg>)
 a product with an odd number of negative factors
 a sum of two or more negative terms

pos
 positive numeric constants
 variables
 function calls other than +,*
 a product with an even number of negative factors
 a sum of two or more positive terms?

what about a mixed sum?
(* -1/2 (+ (/ (- a)) b))?

so that the rules don't loop, we muct ensure that a negative term * -1 is not negative


|#



